#include <J3ML/Geometry/AABB.h>
#include <cassert>
#include <J3ML/Geometry/Plane.h>
#include <J3ML/Geometry/Sphere.h>
//#include <J3ML/Geometry/OBB.h>
#include <J3ML/Geometry/LineSegment.h>
#include <J3ML/Geometry/Triangle.h>
#include <J3ML/Geometry/Polygon.h>
#include <J3ML/Geometry/Frustum.h>
#include <J3ML/Geometry/Capsule.h>
#include <J3ML/Geometry/Ray.h>
#include <J3ML/Geometry/TriangleMesh.h>
#include <J3ML/Geometry/Polyhedron.h>
#include <J3ML/Algorithm/RNG.h>

namespace J3ML::Geometry {

    /// See Christer Ericson's Real-time Collision Detection, p. 87, or
    /// James Arvo's "Transforming Axis-aligned Bounding Boxes" in Graphics Gems 1, pp. 548-550.
    /// http://www.graphicsgems.org/
    template<typename Matrix>
    void AABBTransformAsAABB(AABB &aabb, Matrix &m)
    {
        const Vector3 centerPoint = (aabb.minPoint + aabb.maxPoint) * 0.5f;
        const Vector3 halfSize = centerPoint - aabb.minPoint;
        Vector3 newCenter = m.Mul(centerPoint);


        // The following is equal to taking the absolute value of the whole matrix m.
        Vector3 newDir = Vector3(std::abs(m[0][0] * halfSize.x) + std::abs(m[0][1] * halfSize.y) + std::abs(m[0][2] * halfSize.z),
                                 std::abs(m[1][0] * halfSize.x) + std::abs(m[1][1] * halfSize.y) + std::abs(m[1][2] * halfSize.z),
                                 std::abs(m[2][0] * halfSize.x) + std::abs(m[2][1] * halfSize.y) + std::abs(m[2][2] * halfSize.z));
        aabb.minPoint = newCenter - newDir;
        aabb.maxPoint = newCenter + newDir;
    }

    AABB AABB::FromCenterAndSize(const J3ML::Geometry::Vector3 &center, const J3ML::Geometry::Vector3 &size) {
        Vector3 halfSize =  size * 0.5f;
        return AABB{center - halfSize, center + halfSize};
    }

    float AABB::MinX() const { return minPoint.x; }

    float AABB::MinY() const { return minPoint.y; }

    float AABB::MinZ() const { return minPoint.z; }

    float AABB::MaxX() const { return maxPoint.x; }

    float AABB::MaxY() const { return maxPoint.y; }

    float AABB::MaxZ() const { return maxPoint.z; }

    Sphere AABB::MinimalEnclosingSphere() const {
        return Sphere(Centroid(), Size().Length()*0.5f);
    }

    Vector3 AABB::HalfSize() const {
        return this->Size()/2.f;
    }

    Sphere AABB::MaximalContainedSphere() const {
        Vector3 halfSize = HalfSize();
        return Sphere(Centroid(), std::min(halfSize.x, std::min(halfSize.y, halfSize.z)));
    }

    bool AABB::IsFinite() const {
        return minPoint.IsFinite() && maxPoint.IsFinite();
    }

    Vector3 AABB::Centroid() const {
        return (minPoint+maxPoint) * 0.5f;
    }

    Vector3 AABB::Size() const {
        return this->maxPoint - this->minPoint;
    }

    Vector3 AABB::PointInside(float x, float y, float z) const {
        Vector3 d = maxPoint - minPoint;
        return minPoint + d.Mul({x, y, z});
    }

    LineSegment AABB::Edge(int edgeIndex) const {
        switch(edgeIndex)
        {
            default:
            case 0: return LineSegment(minPoint, {minPoint.x, minPoint.y, maxPoint.z});
        }
    }

    Vector3 AABB::CornerPoint(int cornerIndex) const {
        // TODO: assert(0 <= cornerIndex && cornerIndex <= 7)
        switch(cornerIndex)
        {
            default:
            case 0: return minPoint;
            case 1: return {minPoint.x, minPoint.y, maxPoint.z};
            case 2: return {minPoint.x, maxPoint.y, minPoint.z};
            case 3: return {minPoint.x, maxPoint.y, maxPoint.z};
            case 4: return {maxPoint.x, minPoint.y, minPoint.z};
            case 5: return {maxPoint.x, minPoint.y, maxPoint.z};
            case 6: return {maxPoint.x, maxPoint.y, minPoint.z};
            case 7: return maxPoint;
        }
    }

    Vector3 AABB::ExtremePoint(const Vector3 &direction) const {
        return {direction.x >= 0.f ? maxPoint.x : minPoint.x,
                direction.y >= 0.f ? maxPoint.y : minPoint.y,
                direction.z >= 0.f ? maxPoint.z : minPoint.z};
    }

    Vector3 AABB::ExtremePoint(const Vector3 &direction, float &projectionDistance) const {
        auto extremePt = ExtremePoint(direction);
        projectionDistance = extremePt.Dot(direction);
        return extremePt;
    }

    Vector3 AABB::PointOnEdge(int edgeIndex, float u) const {
        // TODO: assert(0 <= edgeIndex && edgeIndex <= 11);
        // TODO: assert(0 <= u && u < 1.f);

        auto d = maxPoint - minPoint;
        switch(edgeIndex) {
            default:
            case 0: return {minPoint.x, minPoint.y, minPoint.z + u * d.z};
            case 1: return {minPoint.x, maxPoint.y, minPoint.z + u * d.z};
            case 2: return {maxPoint.x, minPoint.y, minPoint.z + u * d.z};
            case 3: return {maxPoint.x, maxPoint.y, minPoint.z + u * d.z};

            case 4: return {minPoint.x, minPoint.y + u * d.y, minPoint.z};
            case 5: return {maxPoint.x, minPoint.y + u * d.y, minPoint.z};
            case 6: return {minPoint.x, minPoint.y + u * d.y, maxPoint.z};
            case 7: return {maxPoint.x, minPoint.y + u * d.y, maxPoint.z};

            case 8: return {minPoint.x + u * d.x, minPoint.y, minPoint.z};
            case 9: return {minPoint.x + u * d.x, minPoint.y, maxPoint.z};
            case 10:return {minPoint.x + u * d.x, maxPoint.y, minPoint.z};
            case 11:return {minPoint.x + u * d.x, maxPoint.y, maxPoint.z};
        }
    }

    Vector3 AABB::FaceCenterPoint(int faceIndex) const {
        // TODO: assert(0 <= faceIndex && faceIndex <= 5)
        auto center = (minPoint + maxPoint) * 0.5f;
        switch (faceIndex) {
            default:
            case 0: return {minPoint.x, center.y, center.z};
            case 1: return {maxPoint.x, center.y, center.z};
            case 2: return {center.x, minPoint.y, center.z};
            case 3: return {center.x, maxPoint.y, center.z};
            case 4: return {center.x, center.y, minPoint.z};
            case 5: return {center.x, center.y, maxPoint.z};
        }
    }

    Vector3 AABB::FacePoint(int faceIndex, float u, float v) const {
        // TODO: assert(0 <= faceIndex && faceIndex <= 5);
        // TODO: assert(0 <= u && u <= 1.f);
        // TODO: assert(0 <= v && v <= 1.f);

        auto d = maxPoint - minPoint;
        switch(faceIndex)
        {
            default: // For release builds where assume() is disabled, return always the first option if out-of-bounds.
            case 0: return {minPoint.x, minPoint.y + u * d.y, minPoint.z + v * d.z};
            case 1: return {maxPoint.x, minPoint.y + u * d.y, minPoint.z + v * d.z};
            case 2: return {minPoint.x + u * d.x, minPoint.y, minPoint.z + v * d.z};
            case 3: return {minPoint.x + u * d.x, maxPoint.y, minPoint.z + v * d.z};
            case 4: return {minPoint.x + u * d.x, minPoint.y + v * d.y, minPoint.z};
            case 5: return {minPoint.x + u * d.x, minPoint.y + v * d.y, maxPoint.z};
        }
    }

    Vector3 AABB::FaceNormal(int faceIndex) const {
        // TODO: assert(0 <= faceIndex && faceIndex <= 5);
        switch(faceIndex) {
            default:
            case 0: return {-1,  0,  0};
            case 1: return { 1,  0,  0};
            case 2: return { 0, -1,  0};
            case 3: return { 0,  1,  0};
            case 4: return { 0,  0, -1};
            case 5: return { 0,  0,  1};
        }
    }

    Plane AABB::FacePlane(int faceIndex) const {
        // TODO: assert(0 <= faceIndex && faceIndex <= 5);
        return Plane(FaceCenterPoint(faceIndex), FaceNormal(faceIndex));
    }

    AABB AABB::MinimalEnclosingAABB(const Vector3 *pointArray, int numPoints) {
        AABB aabb;
        aabb.SetFrom(pointArray, numPoints);
        return aabb;
    }

    float AABB::Volume() const {
        Vector3 sz = Size();
        return sz.x * sz.y * sz.z;
    }

    float AABB::SurfaceArea() const {
        Vector3 size = Size();
        return 2.f * (size.x*size.y + size.x*size.z + size.y*size.z);
    }

    void AABB::SetFromCenterAndSize(const Vector3& center, const Vector3& size)
    {
        Vector3 halfSize = 0.5f * size;
        minPoint = center - halfSize;
        maxPoint = center + halfSize;
    }


    void AABB::SetFrom(const OBB& obb)
    {
        Vector3 halfSize = Vector3::Abs(obb.axis[0] * obb.r[0]) + Vector3::Abs(obb.axis[1]*obb.r[1]) + Vector3::Abs(obb.axis[2]*obb.r[2]);
        SetFromCenterAndSize(obb.pos, 2.f*halfSize);
    }

    void AABB::SetFrom(const Sphere& s)
    {
        Vector3 d = Vector3(s.Radius, s.Radius, s.Radius);
        minPoint = s.Position - d;
        maxPoint = s.Position + d;
    }

    void AABB::SetFrom(const Vector3 *pointArray, int numPoints) {
        assert(pointArray || numPoints == 0);
        SetNegativeInfinity();
        if (!pointArray)
            return;
        for (int i = 0; i < numPoints; ++i)
            Enclose(pointArray[i]);
    }

    Vector3 AABB::GetRandomPointInside(J3ML::Algorithm::RNG &rng) const {
        float f1 = rng.Float();
        float f2 = rng.Float();
        float f3 = rng.Float();

        return PointInside(f1, f2, f3);
    }

    void AABB::SetNegativeInfinity() {
        minPoint = Vector3::Infinity;
        maxPoint = Vector3::NegativeInfinity;
    }

    void AABB::Enclose(const Vector3& point) {
        minPoint = Vector3::Min(minPoint, point);
        maxPoint = Vector3::Max(maxPoint, point);
    }

    void AABB::Enclose(const Vector3& aabbMinPt, const Vector3& aabbMaxPt)
    {
        minPoint = Vector3::Min(minPoint, aabbMinPt);
        maxPoint = Vector3::Max(maxPoint, aabbMaxPt);
    }

    void AABB::Enclose(const LineSegment& lineSegment)
    {
       Enclose(Vector3::Min(lineSegment.A, lineSegment.B), Vector3::Max(lineSegment.A, lineSegment.B));
    }

    void AABB::Enclose(const OBB& obb)
    {
        Vector3 absAxis0 = obb.axis[0].Abs();
        Vector3 absAxis1 = obb.axis[1].Abs();
        Vector3 absAxis2 = obb.axis[2].Abs();
        Vector3 d = obb.r.x * absAxis0 + obb.r.y * absAxis1 + obb.r.z * absAxis2;
    }

    Vector3 AABB::GetClosestPoint(const Vector3 &point) const {
        Vector3 result = point;
        if (point.x > this->maxPoint.x)
            result.x = this->maxPoint.x;
        else if (point.x < this->minPoint.x)
            result.x = this->minPoint.x;
        else
            result.x = point.x;

        if (point.y > this->maxPoint.y)
            result.y = this->maxPoint.y;
        else if (point.y < this->minPoint.y)
            result.y = this->minPoint.y;
        else
            result.y = point.y;

        if (point.z > this->maxPoint.z)
            result.z = this->maxPoint.z;
        else if (point.z < this->minPoint.z)
            result.z = this->minPoint.z;
        else
            result.z = point.z;
    }

    AABB::AABB(const Vector3 &min, const Vector3 &max) : Shape(), minPoint(min), maxPoint(max)
    {

    }

    AABB::AABB() : Shape() {}

    float Max(float a, float b)
    {
        return std::max(a, b);
    }
    float Max(float a, float b, float c)
    {
        return std::max(a, std::max(b, c));
    }

    float Min(float a, float b, float c)
    {
        return std::min(a, std::min(b, c));
    }

    // Compute the face normals of the AABB, because the AABB is at center
    // and (of course) axis aligned, we know it's normals are the X,Y,Z axes.
    Vector3 u0 = Vector3(1.f, 0.f, 0.f);
    Vector3 u1 = Vector3(0.f, 1.f, 0.f);
    Vector3 u2 = Vector3(0.f, 0.f, 1.f);





    bool AABB::TestAxis(const Vector3& axis, const Vector3& v0, const Vector3& v1, const Vector3& v2) const
    {

        Vector3 e = this->Size();

        // Testing axis: axis_u0_f0
        // Project all 3 vertices of the triangle onto the Separating axis
        float p0 = Vector3::Dot(v0, axis);
        float p1 = Vector3::Dot(v1, axis);
        float p2 = Vector3::Dot(v2, axis);

        // Project the AABB onto the separating axis
        // We don't care about the end points of the projection
        // just the length of the half-size of the AABB
        // that is, we're only casting the extents onto the
        // separating axis, not the AABB center. We don't
        // need to cast the center, because we know that the
        // AABB is at origin compared to the triangle!
        float r = e.x * std::abs(Vector3::Dot(u0, axis)) +
                  e.y * std::abs(Vector3::Dot(u1, axis)) +
                  e.z * std::abs(Vector3::Dot(u2, axis));

        // Now do the actual test, basically see if either of
        // the most extreme of the triangle points intersects r
        // You might need to write Min & Max functions that take 3 arguments
        if (Max(Max(p0, p1, p2), Min(p0, p1, p2)) > r)
        {
            // This means BOTH of the points of the projected triangle
            // are outside the projected half-length of the AABB
            // Therefore the axis is separating and we can exit
            return false;
        }

        return true;
    }


    bool AABB::Intersects(const Triangle &triangle) const {
        // https://gdbooks.gitbooks.io/3dcollisions/content/Chapter4/aabb-triangle.html

        Vector3 v0 = triangle.V0;
        Vector3 v1 = triangle.V1;
        Vector3 v2 = triangle.V2;

        // Convert AABB to center-extentss form
        Vector3 c = this->Centroid();
        Vector3 e = this->Size();

        // Translate the triangle as conceptually moving the AABB to origin
        // This is the same as we did with the point in triangle test
        v0 -= c;
        v1 -= c;
        v2 -= c;

        // Compute the edge vectors of the triangle
        // That is , get the lines between the points as vectors
        Vector3 f0 = v1 - v0; // B - A
        Vector3 f1 = v2 - v1; // C - B
        Vector3 f2 = v0 - v2; // A - C

        // There are a total of 13 axes to test!!!
        // We first test against 9 axis, these axes are given by cross product combinations
        // of the edges of the triangle and the edges of the AABB. You need to get an axis testing each of the 3 sides
        // of the AABB against each of the 3 sides of the triangle. The result is 9 axes of separation.

        // Compute the 9 axes
        Vector3 axis_u0_f0 = Vector3::Cross(u0, f0);
        Vector3 axis_u0_f1 = Vector3::Cross(u0, f1);
        Vector3 axis_u0_f2 = Vector3::Cross(u0, f2);
        Vector3 axis_u1_f0 = Vector3::Cross(u1, f0);
        Vector3 axis_u1_f1 = Vector3::Cross(u1, f1);
        Vector3 axis_u1_f2 = Vector3::Cross(u1, f2);
        Vector3 axis_u2_f0 = Vector3::Cross(u1, f0);
        Vector3 axis_u2_f1 = Vector3::Cross(u1, f1);
        Vector3 axis_u2_f2 = Vector3::Cross(u1, f2);

        if (TestAxis(axis_u0_f0, v0, v1, v2)) return true;
        if (TestAxis(axis_u0_f1, v0, v1, v2)) return true;
        if (TestAxis(axis_u0_f2, v0, v1, v2)) return true;
        if (TestAxis(axis_u1_f0, v0, v1, v2)) return true;
        if (TestAxis(axis_u1_f1, v0, v1, v2)) return true;
        if (TestAxis(axis_u1_f2, v0, v1, v2)) return true;
        if (TestAxis(axis_u2_f0, v0, v1, v2)) return true;
        if (TestAxis(axis_u2_f1, v0, v1, v2)) return true;
        if (TestAxis(axis_u2_f2, v0, v1, v2)) return true;

        // Next we have 3 face normals from the AABB
        // for these tests we are conceptually checking if the bounding box
        // of the triangle intersects the bounding box of the AABB
        // that is to say, the separating axis for all tests are axis aligned:
        // axis1: (1, 0, 0), axis2: (0, 1, 0), axis3: (0, 0, 1)
        // Do the SAT given the 3 primary axes of the AABB
        // You already have two vectors for this: u0, u1, and u2

        if (TestAxis(u0, v0, v1, v2)) return true;
        if (TestAxis(u1, v0, v1, v2)) return true;
        if (TestAxis(u2, v0, v1, v2)) return true;

        // Finally we have one last axis to test, the face normal of the triangle
        // We can get the normal of the triangle by crossing the first two line segments
        Vector3 triangleNormal = Vector3::Cross(f0, f1);
        if (TestAxis(triangleNormal, u0, u1, u2))
            return true;

        // Passed testing for all 13 separating axes that exist
        return false;
    }

    bool AABB::Intersects(const LineSegment &lineSegment) const
    {
        Vector3 dir = lineSegment.B - lineSegment.A;
        float len = dir.Length();
        if (len <= 1e-4f) // Degenerate line segment? Fall back to point-in-AABB test.
            return Contains(lineSegment.A);

        float invLen = 1.f / len;
        dir *= invLen;
        float tNear = 0.f, tFar = len;
#ifdef MATH_SIMD
        return IntersectLineAABB_SSE(lineSegment.a, dir, tNear, tFar);
#else
        return IntersectLineAABB_CPP(lineSegment.A, dir, tNear, tFar);
#endif
    }

    bool AABB::Intersects(const AABB& aabb) const {
        return Intersection(aabb).has_value();
    }

    std::optional<AABB> AABB::Intersection(const AABB& rhs) const {
        // Here we do SAT, except that due to both objects being AABBs, they are "already projected" onto the same axis
        constexpr auto test = [](float a, float b, float c, float d) -> std::optional<Vector2> {
            // Overlap Test
            // Points go:
            //       +-------------+
            // +-----|-----+       |
            // |  1  |     |     2 |
            // |     +-----|-------+
            // +-----------+
            //
            // A-----C-----B-------D
            //
            // IF A < C AND B > C ( Overlap in order object 1 -> object 2)
            // IF C < A AND D > A ( Overlap in order object 2 -> object 1)
            if (a < c && b > c) {
                return Vector2{c, b};
            }
            if (c < a && d > a) {
                return Vector2{a, d};
            }
            return std::nullopt;
        };

        // This is SAT, so we need all axes to collide
        std::optional<Vector2> xCollision = test(MinX(), MaxX(), rhs.MinX(), rhs.MaxX());
        if (!xCollision.has_value()) return std::nullopt;

        std::optional<Vector2> yCollision = test(MinY(), MaxY(), rhs.MinY(), rhs.MaxY());
        if (!yCollision.has_value()) return std::nullopt;

        std::optional<Vector2> zCollision = test(MinZ(), MaxZ(), rhs.MinZ(), rhs.MaxZ());
        if (!zCollision.has_value()) return std::nullopt;

        // At this point all 3 optionals have a value ; x of each is the "min" value, y of each is the "max" value
        return AABB{
            Vector3{xCollision->x, yCollision->x, zCollision->x},
            Vector3{xCollision->y, yCollision->y, zCollision->y}
        };
    }

    bool AABB::IntersectLineAABB_CPP(const Vector3 &linePos, const Vector3 &lineDir, float &tNear, float &tFar) const
    {
        assert(lineDir.IsNormalized());
        assert(tNear <= tFar && "AABB::IntersectLineAABB: User gave a degenerate line as input for the intersection test!");
        // The user should have inputted values for tNear and tFar to specify the desired subrange [tNear, tFar] of the line
        // for this intersection test.
        // For a Line-AABB test, pass in
        //    tNear = -FLOAT_INF;
        //    tFar = FLOAT_INF;
        // For a Ray-AABB test, pass in
        //    tNear = 0.f;
        //    tFar = FLOAT_INF;
        // For a LineSegment-AABB test, pass in
        //    tNear = 0.f;
        //    tFar = LineSegment.Length();

        // Test each cardinal plane (X, Y and Z) in turn.
        if (!Math::EqualAbs(lineDir.x, 0.f))
        {
            float recipDir = Math::RecipFast(lineDir.x);
            float t1 = (minPoint.x - linePos.x) * recipDir;
            float t2 = (maxPoint.x - linePos.x) * recipDir;

            // tNear tracks distance to intersect (enter) the AABB.
            // tFar tracks the distance to exit the AABB.
            if (t1 < t2)
                tNear = Max(t1, tNear), tFar = std::min(t2, tFar);
            else // Swap t1 and t2.
                tNear = Max(t2, tNear), tFar = std::min(t1, tFar);

            if (tNear > tFar)
                return false; // Box is missed since we "exit" before entering it.
        }
        else if (linePos.x < minPoint.x || linePos.x > maxPoint.x)
            return false; // The ray can't possibly enter the box, abort.

        if (!Math::EqualAbs(lineDir.y, 0.f))
        {
            float recipDir = Math::RecipFast(lineDir.y);
            float t1 = (minPoint.y - linePos.y) * recipDir;
            float t2 = (maxPoint.y - linePos.y) * recipDir;

            if (t1 < t2)
                tNear = Max(t1, tNear), tFar = std::min(t2, tFar);
            else // Swap t1 and t2.
                tNear = Max(t2, tNear), tFar = std::min(t1, tFar);

            if (tNear > tFar)
                return false; // Box is missed since we "exit" before entering it.
        }
        else if (linePos.y < minPoint.y || linePos.y > maxPoint.y)
            return false; // The ray can't possibly enter the box, abort.

        if (!Math::EqualAbs(lineDir.z, 0.f)) // ray is parallel to plane in question
        {
            float recipDir = Math::RecipFast(lineDir.z);
            float t1 = (minPoint.z - linePos.z) * recipDir;
            float t2 = (maxPoint.z - linePos.z) * recipDir;

            if (t1 < t2)
                tNear = Max(t1, tNear), tFar = std::min(t2, tFar);
            else // Swap t1 and t2.
                tNear = Max(t2, tNear), tFar = std::min(t1, tFar);
        }
        else if (linePos.z < minPoint.z || linePos.z > maxPoint.z)
            return false; // The ray can't possibly enter the box, abort.

        return tNear <= tFar;
    }

    Vector3 AABB::AnyPointFast() const { return minPoint;}

    Vector3 AABB::GetRandomPointOnSurface(RNG &rng) const {
        int i = rng.Int(0, 5);
        float f1 = rng.Float();
        float f2 = rng.Float();
        return FacePoint(i, f1, f2);
    }

    Vector3 AABB::GetRandomPointOnEdge(RNG &rng) const {
        int i = rng.Int(0, 11);
        float f = rng.Float();
        return PointOnEdge(i, f);
    }

    Vector3 AABB::GetRandomCornerPoint(RNG &rng) const {
        return CornerPoint(rng.Int(0, 7));
    }

    void AABB::Translate(const Vector3 &offset) {
        minPoint += offset;
        maxPoint += offset;
    }

    AABB AABB::Translated(const Vector3 &offset) const {
        return AABB(minPoint+offset, maxPoint+offset);
    }

    void AABB::Scale(const Vector3 &scale) {
        minPoint.x *= scale.x;
        minPoint.y *= scale.y;
        minPoint.z *= scale.z;
        maxPoint.x *= scale.x;
        maxPoint.y *= scale.y;
        maxPoint.z *= scale.z;
    }

    AABB AABB::Scaled(const Vector3 &scale) const {
        return AABB(
            Vector3(minPoint.x*scale.y, minPoint.y*scale.y, minPoint.z*scale.z),
            Vector3(maxPoint.x*scale.y, maxPoint.y*scale.y, maxPoint.z*scale.z)
        );
    }

    AABB AABB::TransformAABB(const Matrix3x3 &transform) {
        // TODO: assert(transform.IsColOrthogonal());
        // TODO: assert(transform.HasUniformScale());
        AABBTransformAsAABB(*this, transform);
    }

    AABB AABB::TransformAABB(const Matrix4x4 &transform) {
        // TODO: assert(transform.IsColOrthogonal());
        // TODO: assert(transform.HasUniformScale());
        // TODO: assert(transform.Row(3).Equals(0,0,0,1));
        AABBTransformAsAABB(*this, transform);
    }

    AABB AABB::TransformAABB(const Quaternion &transform) {
        Vector3 newCenter = transform.Transform(Centroid());
        Vector3 newDir = Vector3::Abs((transform.Transform(Size())*0.5f));
        minPoint = newCenter - newDir;
        maxPoint = newCenter + newDir;
    }

    OBB AABB::Transform(const Matrix3x3 &transform) const {
        OBB obb;
        obb.SetFrom(*this, transform);
        return obb;
    }

    bool AABB::Contains(const Vector3 &aabbMinPoint, const Vector3 &aabbMaxPoint) const {
        return minPoint.x <= aabbMinPoint.x && maxPoint.x >= aabbMaxPoint.x &&
               minPoint.y <= aabbMinPoint.y && maxPoint.y >= aabbMaxPoint.y &&
               minPoint.z <= aabbMinPoint.z && maxPoint.z >= aabbMaxPoint.z;
    }

    bool AABB::Contains(const LineSegment &lineSegment) const {
        return Contains(Vector3::Min(lineSegment.A, lineSegment.B), Vector3::Max(lineSegment.A, lineSegment.B));
    }



    bool AABB::Contains(const Vector3 &point) const {
        return minPoint.x <= point.x && point.x <= maxPoint.x &&
               minPoint.y <= point.y && point.y <= maxPoint.y &&
               minPoint.z <= point.z && point.z <= maxPoint.z;
    }

    OBB AABB::Transform(const Matrix4x4 &transform) const {
        OBB obb;
        obb.SetFrom(*this, transform);
        return obb;
    }

    OBB AABB::Transform(const Quaternion &transform) const {
        OBB obb;
        obb.SetFrom(*this, transform);
        return obb;
    }

    bool AABB::Contains(const AABB &aabb) const {
        return Contains(aabb.minPoint, aabb.maxPoint);
    }

    bool AABB::Contains(const OBB &obb) const {
        return Contains(obb.MinimalEnclosingAABB());
    }

    bool AABB::Contains(const Sphere &sphere) const {
        auto radiusVec = Vector3(sphere.Radius,sphere.Radius, sphere.Radius);
        return Contains(sphere.Position - radiusVec, sphere.Position + radiusVec);
    }

    bool AABB::Contains(const Capsule &capsule) const {
        return Contains(capsule.MinimalEnclosingAABB());
    }

    bool AABB::Contains(const Triangle &triangle) const {
        return Contains(triangle.BoundingAABB());
    }

    bool AABB::Contains(const Polygon &polygon) const {
        return Contains(polygon.MinimalEnclosingAABB());
    }

    bool AABB::Contains(const Frustum &frustum) const {
        return Contains(frustum.MinimalEnclosingAABB());
    }

    bool AABB::Contains(const Polyhedron &polyhedron) const {
        return Contains(polyhedron.MinimalEnclosingAABB());
    }

    bool AABB::IntersectLineAABB(const Vector3 &linePos, const Vector3 &lineDir, float tNear, float tFar) const {
        //assert(lineDir.IsNormalized() && lineDir && lineDir.LengthSquared());
        assert(tNear <= tFar && "");
        // The user should have inputted values for tNear and tFar to specify the desired subrange [tNear, tFar] of the line
        // for this intersection test.
        // For a Line-AABB test, pass in
        //    tNear = -FLOAT_INF;
        //    tFar = FLOAT_INF;
        // For a Ray-AABB test, pass in
        //    tNear = 0.f;
        //    tFar = FLOAT_INF;
        // For a LineSegment-AABB test, pass in
        //    tNear = 0.f;
        //    tFar = LineSegment.Length();

        // Test each cardinal plane (X, Y, and Z) in turn.
        if (!Math::EqualAbs(lineDir.x, 0.f)) {
            float recipDir = 1.f / lineDir.x;
            float t1 = (minPoint.x - linePos.x) * recipDir;
            float t2 = (maxPoint.x - linePos.x) * recipDir;

            // tNear tracks distance to intersect (enter) the AABB
            // tFar tracks the distance to exit the AABB
            if (t1 < t2)
                tNear = std::max(t1, tNear), tFar = std::min(t2, tFar);
            else // swap t1 and t2;
                tNear = std::max(t2, tNear), tFar = std::min(t1, tFar);

            if (tNear > tFar)
                return false; // Box is missed since we "exit" before entering it
        }
        else if (linePos.x < minPoint.x || linePos.x > maxPoint.x)
            return false; // the ray can't possibly enter the box

        if (!Math::EqualAbs(lineDir.y, 0.f)) // ray is parallel to plane in question
        {
            float recipDir = 1.f / lineDir.y;
            float t1 = (minPoint.y - linePos.y) * recipDir;
            float t2 = (maxPoint.y - linePos.y) * recipDir;

            if (t1 < t2)
                tNear = std::max(t1, tNear), tFar = std::min(t2, tFar);
            else
                tNear = std::max(t2, tNear), tFar = std::min(t1, tFar);

            if (tNear > tFar)
                return false;
        }
        else if (linePos.y < minPoint.y || linePos.y > maxPoint.y)
            return false; // The ray can't possibly enter the box, abort.

        if (!Math::EqualAbs(lineDir.z, 0.f)) // ray is parallel to plane in question
        {
            float recipDir = 1.f / lineDir.z;
            float t1 = (minPoint.z - linePos.z) * recipDir;
            float t2 = (maxPoint.z - linePos.z) * recipDir;

            if (t1 < t2)
                tNear = std::max(t1, tNear), tFar = std::min(t2, tFar);
            else // Swap t1 and t2.
                tNear = std::max(t2, tNear), tFar = std::min(t1, tFar);
        } else if (linePos.z < minPoint.z || linePos.z > maxPoint.z)
            return false;

        return tNear <= tFar;

    }

    void AABB::ProjectToAxis(const Vector3 &axis, float &dMin, float &dMax) const {
        Vector3 c = (minPoint + maxPoint) * 0.5f;
        Vector3 e = maxPoint - c;

        // Compute the projection interval radius of the AABB onto L(t) = aabb.center + t * plane.normal;
        float r = std::abs(e[0]*std::abs(axis[0]) + e[1]*std::abs(axis[1]) + e[2]*std::abs(axis[2]));

        // Compute the distance of the box center from plane.
        float s = axis.Dot(c);
        dMin = s - r;
        dMax = s + r;
    }

}
