#include <J3ML/Geometry/Capsule.h>
#include <J3ML/Geometry/AABB.h>
#include <J3ML/Geometry/Sphere.h>
#include <J3ML/Geometry/Polygon.h>

namespace J3ML::Geometry
{

    Capsule::Capsule() : l() {}

    AABB Capsule::MinimalEnclosingAABB() const
    {
        Vector3 d = Vector3(r, r, r);
        AABB aabb(Vector3::Min(l.A, l.B) - d, Vector3::Max(l.A, l.B) + d);
        return aabb;
    }

    void Capsule::ProjectToAxis(const Vector3 &direction, float &outMin, float &outMax) const {
        outMin = Vector3::Dot(direction, l.A);
        outMax = Vector3::Dot(direction, l.B);
        if (outMax < outMin)
            Swap(outMin, outMax);

        // The following requires that direction is normalized, otherwise we would have to sub/add 'r * direction.Length()', but
        // don't want to do that for performance reasons.
        assert(direction.IsNormalized());
        outMin -= r;
        outMax += r;
    }

    bool Capsule::Intersects(const Ray &ray) const
    {
        return l.Distance(ray) <= r;
    }

    bool Capsule::Intersects(const Line &line) const
    {
        return l.Distance(line) <= r;
    }

    bool Capsule::Intersects(const LineSegment &lineSegment) const
    {
        return l.Distance(lineSegment) <= r;
    }

    bool Capsule::Intersects(const Plane &plane) const {
        return l.Distance(plane) <= r;
    }

    bool Capsule::Intersects(const AABB &aabb) const
    {
        //return FloatingPointOffsetedGJKIntersect(*this, aabb);
    }

    bool Capsule::Intersects(const OBB &obb) const
    {
        //return GJKIntersect(*this, obb);
    }

/// [groupSyntax]
    bool Capsule::Intersects(const Sphere &sphere) const
    {
        float R = r + sphere.Radius;
        return l.DistanceSq(sphere.Position) <= R*R;
    }

/// [groupSyntax]
    bool Capsule::Intersects(const Capsule &capsule) const
    {
        float R = r + capsule.r;
        return l.DistanceSq(capsule.l) <= R*R;
    }

    bool Capsule::Intersects(const Triangle &triangle) const
    {
        Vector3 thisPoint;
        Vector3 trianglePoint = triangle.ClosestPoint(l, &thisPoint);
        return thisPoint.DistanceSq(trianglePoint) <= r*r;
    }

    bool Capsule::Intersects(const Polygon &polygon) const
    {
        return polygon.Intersects(*this);
    }

    bool Capsule::Intersects(const Frustum &frustum) const
    {
        return frustum.Intersects(*this);
    }

    bool Capsule::Intersects(const Polyhedron &polyhedron) const
    {
        return polyhedron.Intersects(*this);
    }

    Vector3 Capsule::ExtremePoint(const Vector3 &direction) const {
        float len = direction.Length();
        assert(len > 0.f);
        return (Vector3::Dot(direction, l.B - l.A) >= 0.f ? l.B : l.A) + direction * (r / len);
    }

    Vector3 Capsule::ExtremePoint(const Vector3 &direction, float &projectionDistance) const {
        Vector3 extremePoint = ExtremePoint(direction);
        projectionDistance = extremePoint.Dot(direction);
        return extremePoint;
    }
}