#include <J3ML/Geometry/Common.h>
#include <J3ML/Geometry/Frustum.h>
#include <cmath>
#include "J3ML/Geometry/AABB.h"
#include <J3ML/Geometry/Polyhedron.h>
#include <J3ML/Geometry/LineSegment.h>
#include <J3ML/Geometry/Polygon.h>
#include <J3ML/Algorithm/GJK.h>


namespace J3ML::Geometry
{

    void Frustum::GetCornerPoints(Vector3 *outPointArray) const
    {
        assert(outPointArray);

        if (type == FrustumType::Perspective)
        {
            float tanhfov = std::tan(horizontalFov*0.5f);
            float tanvfov = std::tan(verticalFov*0.5f);
            float frontPlaneHalfWidth = tanhfov*nearPlaneDistance;
            float frontPlaneHalfHeight = tanvfov*nearPlaneDistance;
            float farPlaneHalfWidth = tanhfov*farPlaneDistance;
            float farPlaneHalfHeight = tanvfov*farPlaneDistance;

            Vector3 right = WorldRight();

            Vector3 nearCenter = pos + front * nearPlaneDistance;
            Vector3 nearHalfWidth = frontPlaneHalfWidth*right;
            Vector3 nearHalfHeight = frontPlaneHalfHeight*up;
            outPointArray[0] = nearCenter - nearHalfWidth - nearHalfHeight;
            outPointArray[1] = nearCenter + nearHalfWidth - nearHalfHeight;
            outPointArray[2] = nearCenter - nearHalfWidth + nearHalfHeight;
            outPointArray[3] = nearCenter + nearHalfWidth + nearHalfHeight;

            Vector3 farCenter = pos + front * farPlaneDistance;
            Vector3 farHalfWidth = farPlaneHalfWidth*right;
            Vector3 farHalfHeight = farPlaneHalfHeight*up;
            outPointArray[4] = farCenter - farHalfWidth - farHalfHeight;
            outPointArray[5] = farCenter + farHalfWidth - farHalfHeight;
            outPointArray[6] = farCenter - farHalfWidth + farHalfHeight;
            outPointArray[7] = farCenter + farHalfWidth + farHalfHeight;
        }
        else
        {
            Vector3 right = WorldRight();
            Vector3 nearCenter = pos + front * nearPlaneDistance;
            Vector3 farCenter = pos + front * farPlaneDistance;
            Vector3 halfWidth = orthographicWidth * 0.5f * right;
            Vector3 halfHeight = orthographicHeight * 0.5f * up;

            outPointArray[0] = nearCenter - halfWidth - halfHeight;
            outPointArray[1] = nearCenter + halfWidth - halfHeight;
            outPointArray[2] = nearCenter - halfWidth + halfHeight;
            outPointArray[3] = nearCenter + halfWidth + halfHeight;
            outPointArray[4] = farCenter - halfWidth - halfHeight;
            outPointArray[5] = farCenter + halfWidth - halfHeight;
            outPointArray[6] = farCenter - halfWidth + halfHeight;
            outPointArray[7] = farCenter + halfWidth + halfHeight;
        }
    }

    void Frustum::ProjectToAxis(const Vector3 &direction, float &outMin, float &outMax) const
    {
        Vector3 corners[8];
        GetCornerPoints(corners);
        outMax = -INFINITY;
        outMin = INFINITY;
        for(int i = 0; i < 8; ++i)
        {
            float d = Vector3::Dot(direction, corners[i]);
            outMax = std::max(outMax, d);
            outMin = std::min(outMin, d);
        }
    }

    LineSegment Frustum::Edge(int edgeIndex) const
    {
        assert(0 <= edgeIndex && edgeIndex <= 11);
        switch(edgeIndex)
        {
            default: // For release builds where assume() is disabled, return always the first option if out-of-bounds.
            case 0: return {CornerPoint(0), CornerPoint(1)};
            case 1: return LineSegment(CornerPoint(0), CornerPoint(2));
            case 2: return LineSegment(CornerPoint(0), CornerPoint(4));
            case 3: return LineSegment(CornerPoint(1), CornerPoint(3));
            case 4: return LineSegment(CornerPoint(1), CornerPoint(5));
            case 5: return LineSegment(CornerPoint(2), CornerPoint(3));
            case 6: return LineSegment(CornerPoint(2), CornerPoint(6));
            case 7: return LineSegment(CornerPoint(3), CornerPoint(7));
            case 8: return LineSegment(CornerPoint(4), CornerPoint(5));
            case 9: return LineSegment(CornerPoint(4), CornerPoint(6));
            case 10: return LineSegment(CornerPoint(5), CornerPoint(7));
            case 11: return LineSegment(CornerPoint(6), CornerPoint(7));
        }
    }

    Vector3 Frustum::ExtremePoint(const Vector3 &direction, float &projectionDistance) const
    {
        Vector3 corners[8];
        GetCornerPoints(corners);
        Vector3 mostExtreme = Vector3::NaN;
        projectionDistance = -INFINITY;
        for(int i = 0; i < 8; ++i)
        {
            float d = Vector3::Dot(direction, corners[i]);
            if (d > projectionDistance)
            {
                projectionDistance = d;
                mostExtreme = corners[i];
            }
        }
        return mostExtreme;
    }

    Frustum Frustum::CreateFrustumFromCamera(const CoordinateFrame &cam, float aspect, float fovY, float zNear, float zFar) {
        Frustum frustum;
        const float halfVSide = zFar * tanf(fovY * 0.5f);
        const float halfHSide = halfVSide * aspect;

        const Vector3 frontMultFar = cam.Front * zFar;

        // frustum.NearFace  = Plane{cam.Position + cam.Front * zNear, cam.Front};
        // frustum.FarFace   = Plane{cam.Position + frontMultFar, -cam.Front};
        // frustum.RightFace = Plane{cam.Position, Vector3::Cross(frontMultFar - cam.Right * halfHSide, cam.Up)};
        // frustum.LeftFace  = Plane{cam.Position, Vector3::Cross(cam.Up, frontMultFar+cam.Right*halfHSide)};
        // frustum.TopFace   = Plane{cam.Position, Vector3::Cross(cam.Right, frontMultFar - cam.Up * halfVSide)};
        // frustum.BottomFace   = Plane{cam.Position, Vector3::Cross(frontMultFar + cam.Up * halfVSide, cam.Right)};
        return frustum;
    }

    AABB Frustum::MinimalEnclosingAABB() const {
        AABB aabb;
        aabb.SetNegativeInfinity();
        for(int i = 0; i < 8; ++i)
            aabb.Enclose(CornerPoint(i));
        return aabb;
    }

    Vector3 Frustum::CornerPoint(int cornerIndex) const {
        assert(0 <= cornerIndex && cornerIndex <= 7);
        switch(cornerIndex)
        {
            default: // For release builds where assume() is disabled, return always the first option if out-of-bounds.
            case 0: return NearPlanePos(-1, -1);
            case 1: return FarPlanePos(-1, -1);
            case 2: return NearPlanePos(-1, 1);
            case 3: return FarPlanePos(-1, 1);
            case 4: return NearPlanePos(1, -1);
            case 5: return FarPlanePos(1, -1);
            case 6: return NearPlanePos(1, 1);
            case 7: return FarPlanePos(1, 1);
        }
    }

    Vector3 Frustum::NearPlanePos(float x, float y) const {
        assert(type == FrustumType::Perspective || type == FrustumType::Orthographic);

        if (type == FrustumType::Perspective)
        {
            float frontPlaneHalfWidth = std::tan(horizontalFov*0.5f)*nearPlaneDistance;
            float frontPlaneHalfHeight = std::tan(verticalFov*0.5f)*nearPlaneDistance;
            x = x * frontPlaneHalfWidth; // Map [-1,1] to [-width/2, width/2].
            y = y * frontPlaneHalfHeight;  // Map [-1,1] to [-height/2, height/2].
            Vector3 right = WorldRight();
            return pos + front * nearPlaneDistance + x * right + y * up;
        }
        else
        {
            Vector3 right = WorldRight();
            return pos + front * nearPlaneDistance
                   + x * orthographicWidth * 0.5f * right
                   + y * orthographicHeight * 0.5f * up;
        }
    }

    Vector3 Frustum::FarPlanePos(float x, float y) const {
        assert(type == FrustumType::Perspective || type == FrustumType::Orthographic);

        if (type == FrustumType::Perspective)
        {
            float farPlaneHalfWidth = std::tan(horizontalFov*0.5f)*farPlaneDistance;
            float farPlaneHalfHeight = std::tan(verticalFov*0.5f)*farPlaneDistance;

            x = x * farPlaneHalfWidth;
            y = y * farPlaneHalfHeight;
            Vector3 right = WorldRight();
            return pos + front * farPlaneDistance + x * right + y * up;
        } else {
            Vector3 right = WorldRight();
            return pos + front * farPlaneDistance
                   + x * orthographicWidth * 0.5f * right
                   + y * orthographicHeight * 0.5f * up;
        }
    }

    Polyhedron Frustum::ToPolyhedron() const {
        // Note to maintainer: this function is an exact copy of AABB::ToPolyhedron() and OBB::ToPolyhedron().

        Polyhedron p;
        // Populate the corners of this Frustum.
        // They will be in the order 0: ---, 1: --+, 2: -+-, 3: -++, 4: +--, 5: +-+, 6: ++-, 7: +++.
        for (int i = 0; i < 8; ++i)
        {
            p.v.push_back(CornerPoint(i));
        }

        // generate the 6 faces of this Frustum. The function Frustum::GetPlane() has a convention of returning
        // the planes in order near, far, left, right, top, bottom, so follow the same convention here.
        const int faces[6][4] =
                {
                        { 0, 4, 6, 2 }, // Z-: near plane
                        { 1, 3, 7, 5 }, // Z+: far plane
                        { 0, 2, 3, 1 }, // X-: left plane
                        { 4, 5, 7, 6 }, // X+: right plane
                        { 7, 3, 2, 6 }, // Y+: top plane
                        { 0, 1, 5, 4 }, // Y-: bottom plane
                };

        for (int f = 0; f < 6; ++f)
        {
            Polyhedron::Face face;
            if (this->handedness == FrustumHandedness::Left)
            {
                for (int v = 0; v < 4; ++v)
                    face.v.push_back(faces[f][3-v]);
            } else {
                for (int v = 0; v < 4; ++v)
                    face.v.push_back(faces[f][v]);
            }
            p.f.push_back(face);
        }
        return p;
    }

    bool Frustum::Intersects(const Ray &ray) const {
        return this->ToPolyhedron().Intersects(ray);
    }

    bool Frustum::Intersects(const Line &line) const
    {
        ///@todo This is a naive test. Implement a faster version.
        return this->ToPolyhedron().Intersects(line);
    }

    bool Frustum::Intersects(const LineSegment &lineSegment) const
    {
        return Algorithms::GJKIntersect(*this, lineSegment);
    }

    bool Frustum::Intersects(const AABB &aabb) const
    {
        return Algorithms::GJKIntersect(*this, aabb);
    }

    bool Frustum::Intersects(const OBB &obb) const
    {
        return Algorithms::GJKIntersect(*this, obb);
    }

    bool Frustum::Intersects(const Plane &plane) const
    {
        return plane.Intersects(*this);
    }

    bool Frustum::Intersects(const Triangle &triangle) const
    {
        return Algorithms::GJKIntersect(*this, triangle);
    }

    bool Frustum::Intersects(const Polygon &polygon) const
    {
        return polygon.Intersects(*this);
    }

    bool Frustum::Intersects(const Sphere &sphere) const
    {
        return Algorithms::GJKIntersect(*this, sphere);
    }

    bool Frustum::Intersects(const Capsule &capsule) const
    {
        return Algorithms::GJKIntersect(*this, capsule);
    }

    bool Frustum::Intersects(const Frustum &frustum) const
    {
        return Algorithms::GJKIntersect(*this, frustum);
    }

    bool Frustum::Intersects(const Polyhedron &polyhedron) const
    {
        return this->ToPolyhedron().Intersects(polyhedron);
    }

    Frustum::Frustum()
            : type(FrustumType::Invalid),
              pos(Vector3::NaN),
              front(Vector3::NaN),
              up(Vector3::NaN),
              nearPlaneDistance(NAN),
              farPlaneDistance(NAN),
              worldMatrix(Matrix4x4::NaN),
              viewProjectionMatrix(Matrix4x4::NaN)
    {
        // For conveniency, allow automatic initialization of the graphics API and handedness in use.
        // If neither of the #defines are set, user must specify per-instance.
    }

    Vector3 Frustum::WorldRight() const {
        if (handedness == FrustumHandedness::Right)
            return Vector3::Cross(front, up);
        else
            return Vector3::Cross(up, front);
    }
}