#include <J3ML/LinearAlgebra.h>
#include <J3ML/Geometry/Line.h>
#include <J3ML/Geometry/LineSegment.h>
namespace J3ML::Geometry
{

    /// Computes the closest point pair on two lines.
    /** The first line is specified by two points start0 and end0. The second line is specified by two points start1 and end1.
        The implementation of this function follows http://paulbourke.net/geometry/lineline3d/ .
        @param v0 The starting point of the first line.
        @param v10 The direction vector of the first line. This can be unnormalized.
        @param v2 The starting point of the second line.
        @param v32 The direction vector of the second line. This can be unnormalized.
        @param d [out] Receives the normalized distance of the closest point along the first line.
        @param d2 [out] Receives the normalized distance of the closest point along the second line.
        @return Returns the closest point on line start0<->end0 to the second line.
        @note This is a low-level utility function. You probably want to use ClosestPoint() or Distance() instead.
        @see ClosestPoint(), Distance(). */
    void Line::ClosestPointLineLine(const Vector3 &v0, const Vector3 &v10, const Vector3 &v2, const Vector3 &v32, float &d, float &d2)
    {
        assert(!v10.IsZero());
        assert(!v32.IsZero());
        Vector3 v02 = v0 - v2;
        float d0232 = v02.Dot(v32);
        float d3210 = v32.Dot(v10);
        float d3232 = v32.Dot(v32);
        assert(d3232 != 0.f); // Don't call with a zero direction vector.
        float d0210 = v02.Dot(v10);
        float d1010 = v10.Dot(v10);
        float denom = d1010*d3232 - d3210*d3210;
        if (denom != 0.f)
            d = (d0232*d3210 - d0210*d3232) / denom;
        else
            d = 0.f;
        d2 = (d0232 + d * d3210) / d3232;
    }

    Vector3 Line::GetPoint(float d) const
    {
        assert(Direction.IsNormalized());
        return Position + d * Direction;
    }

    Vector3 Line::ClosestPoint(const Vector3 &targetPoint, float &d) const
    {
        d = Vector3::Dot(targetPoint - Position, Direction);
        return GetPoint(d);
    }

    Vector3 Line::ClosestPoint(const LineSegment &other, float &d, float &d2) const
    {
        ClosestPointLineLine(Position, Direction, other.A, other.B - other.A, d, d2);
        if (d2 < 0.f)
        {
            d2 = 0.f;
            return ClosestPoint(other.A, d);
        }
        else if (d2 > 1.f)
        {
            d2 = 1.f;
            return ClosestPoint(other.B, d);
        }
        else
            return GetPoint(d);
    }



}