#include <J3ML/Geometry/LineSegment.h>
#include "J3ML/Geometry/Capsule.h"
#include <J3ML/Geometry/Line.h>

namespace J3ML::Geometry {

    LineSegment::LineSegment(const Vector3 &a, const Vector3 &b) : A(a), B(b)
    {

    }

    LineSegment::LineSegment() {}

    void LineSegment::ProjectToAxis(const Vector3 &direction, float &outMin, float &outMax) const
    {
        outMin = Vector3::Dot(direction, A);
        outMax = Vector3::Dot(direction, B);
        if (outMax < outMin)
            Swap(outMin, outMax);
    }

    Vector3 LineSegment::GetPoint(float d) const
    {
        return (1.f - d) * A + d * B;
    }

    float LineSegment::Distance(const Vector3 &point, float &d) const
    {
        /// See Christer Ericson's Real-Time Collision Detection, p.130.
        Vector3 closestPoint = ClosestPoint(point, d);
        return closestPoint.Distance(point);
    }

    float LineSegment::DistanceSq(const Vector3 &point) const
    {
        float d;
        /// See Christer Ericson's Real-Time Collision Detection, p.130.
        Vector3 closestPoint = ClosestPoint(point, d);
        return closestPoint.DistanceSq(point);
    }

    float LineSegment::DistanceSq(const LineSegment &other) const
    {
        float d, d2;
        ClosestPoint(other, d, d2);
        return GetPoint(d).DistanceSq(other.GetPoint(d2));
    }


    float LineSegment::Distance(const Plane &other) const {
        {
            float aDist = other.SignedDistance(A);
            float bDist = other.SignedDistance(B);
            if (aDist * bDist < 0.f)
                return 0.f; // There was an intersection, so the distance is zero.
            return std::min(std::abs(aDist), std::abs(bDist));
        }
    }

    float LineSegment::Distance(const LineSegment &other, float &d) const { float d2; return Distance(other, d, d2); }

    float LineSegment::Distance(const LineSegment &other) const { float d, d2; return Distance(other, d, d2); }

    float LineSegment::Distance(const Line &other, float &d) const { float d2; return Distance(other, d, d2); }

    float LineSegment::Distance(const Line &other) const { float d, d2; return Distance(other, d, d2); }

    float LineSegment::Distance(const Ray &other, float &d) const { float d2; return Distance(other, d, d2); }

    float LineSegment::Distance(const Ray &other) const { float d, d2; return Distance(other, d, d2); }

    float LineSegment::Distance(const Vector3 &point) const { float d; return Distance(point, d); }

    Vector3 LineSegment::ClosestPoint(const Vector3 &point, float &d) const {
        Vector3 dir = B - A;
        d = Clamp01(Vector3::Dot(point - A, dir) / dir.LengthSquared());
        return A + d * dir;
    }

    Vector3 LineSegment::ClosestPoint(const Line &other, float &d, float &d2) const
    {
        Line::ClosestPointLineLine(other.Position, other.Direction, A,B - A, d2, d);
        if (d < 0.f)
        {
            d = 0.f;
            other.ClosestPoint(A, d2);
            return A;
        }
        else if (d > 1.f)
        {
            d = 1.f;
            other.ClosestPoint(B, d2);
            return B;
        }
        else
            return GetPoint(d);
    }

    Vector3 LineSegment::ClosestPoint(const Ray &other, float &d, float &d2) const {
        other.ClosestPoint(*this, d2, d);
        return GetPoint(d);
    }

    Vector3 LineSegment::ClosestPoint(const LineSegment &other, float &d, float &d2) const
    {
        Vector3 dir = B - A;
        Line::ClosestPointLineLine(A, B - A, other.A, other.B - other.A, d, d2);
        if (d >= 0.f && d <= 1.f && d2 >= 0.f && d2 <= 1.f)
            return A + d * dir;
        else if (d >= 0.f && d <= 1.f) // Only d2 is out of bounds.
        {
            Vector3 p;
            if (d2 < 0.f)
            {
                d2 = 0.f;
                p = other.A;
            }
            else
            {
                d2 = 1.f;
                p = other.B;
            }
            return ClosestPoint(p, d);
        }
        else if (d2 >= 0.f && d2 <= 1.f) // Only d is out of bounds.
        {
            Vector3 p;
            if (d < 0.f)
            {
                d = 0.f;
                p = A;
            }
            else
            {
                d = 1.f;
                p = B;
            }

            other.ClosestPoint(p, d2);
            return p;
        }
        else // Both u and u2 are out of bounds.
        {
            Vector3 p;
            if (d < 0.f)
            {
                p = A;
                d = 0.f;
            }
            else
            {
                p = B;
                d = 1.f;
            }

            Vector3 p2;
            if (d2 < 0.f)
            {
                p2 = other.A;
                d2 = 0.f;
            }
            else
            {
                p2 = other.B;
                d2 = 1.f;
            }

            float T, T2;
            Vector3 closestPoint = ClosestPoint(p2, T);
            Vector3 closestPoint2 = other.ClosestPoint(p, T2);

            if (closestPoint.DistanceSq(p2) <= closestPoint2.DistanceSq(p))
            {
                d = T;
                return closestPoint;
            }
            else
            {
                d2 = T2;
                return p;
            }
        }
    }

    float LineSegment::Distance(const Ray &other, float &d, float &d2) const {
        {
            ClosestPoint(other, d, d2);
            return GetPoint(d).Distance(other.GetPoint(d2));
        }

    }

    float LineSegment::Distance(const Line &other, float &d, float &d2) const {
        Vector3 closestPoint2 = other.ClosestPoint(*this, d, d2);
        Vector3 closestPoint = GetPoint(d2);
        return closestPoint.Distance(closestPoint2);
    }

    Vector3 LineSegment::Dir() const {
        return (B - A).Normalize();
    }

    float LineSegment::Length() const {
        return A.Distance(B);
    }

    float LineSegment::LengthSq() const
    {
        return A.DistanceSq(B);
    }

    bool LineSegment::Intersects(const LineSegment &segment) const {
        return false;
    }

    Vector3 LineSegment::ExtremePoint(const Vector3 &direction) const {
        return Vector3::Dot(direction, B-A) >= 0.f ? B : A;
    }

    Vector3 LineSegment::ExtremePoint(const Vector3 &direction, float &projectionDistance) const {
        Vector3 extremePoint = ExtremePoint(direction);
        projectionDistance = extremePoint.Dot(direction);
        return extremePoint;
    }

    bool LineSegment::Contains(const Vector3 &point, float distanceThreshold) const {
        return ClosestPoint(point).DistanceSq(point) <= distanceThreshold;
    }

    Vector3 LineSegment::ClosestPoint(const Vector3 &point) const { float d; return ClosestPoint(point, d); }

    float LineSegment::Distance(const LineSegment &other, float &d, float &d2) const {
        ClosestPoint(other, d, d2);
        return GetPoint(d).Distance(other.GetPoint(d2));
    }

    LineSegment operator*(const Matrix4x4 &transform, const LineSegment &l) {
        return LineSegment(transform.Mul(l.A), transform.Mul(l.B));
    }
}