#include <J3ML/Geometry/Ray.h>
#include <J3ML/Geometry/Sphere.h>


namespace J3ML::Geometry
{

    void Ray::ProjectToAxis(const Vector3 &direction, float &outMin, float &outMax) const
    {
        outMin = outMax = Vector3::Dot(direction, Origin);
        float d = Vector3::Dot(direction, Direction);

        // Most of the time, the projection interval will be a half-infinite range, extending to either -inf or +inf.
        if (d > 1e-4f)
            outMax = INFINITY;
        else if (d < -1e4f)
            outMin = -INFINITY;
    }

    RaycastResult Ray::Cast(const Sphere &target, float maxDistance)
    {
        Vector3 p0 = this->Origin;
        Vector3 d = this->Direction.Normalize();

        Vector3 c = target.Position;
        float r = target.Radius;

        Vector3 e = c - p0;
        // Using Length here would cause floating point error to creep in
        float Esq = Vector3::LengthSquared(e);
        float a = Vector3::Dot(e, d);
        float b = std::sqrt(Esq - (a*a));
        float f = std::sqrt((r*r) - (b*b));

        float t = 0;
        // No collision
        if (r * r - Esq + a * a < 0.f)
        {
            t = -1;
        } else if ( Esq < r*r) {
            t = a + f;
        } else
        {
            t = a - f;
        }

        // TODO: Verify this
        Vector3 intersection = p0.Project(d*t);
        Vector3 intersection_from_sphere_origin = (intersection - target.Position);

        Vector3 normal = -intersection_from_sphere_origin.Normalize();

        return RaycastResult{
            intersection,
            normal,
            true,
            (Shape *) &target
        };
    }

    RaycastResult Ray::Cast(const AABB &target, float maxDistance) {
        float t1 = (target.minPoint.x - Origin.x) / Direction.x;
        float t2 = (target.maxPoint.x - Origin.x) / Direction.x;
        float t3 = (target.minPoint.y - Origin.y) / Direction.y;
        float t4 = (target.maxPoint.y - Origin.y) / Direction.y;
        float t5 = (target.minPoint.z - Origin.z) / Direction.z;
        float t6 = (target.maxPoint.z - Origin.z) / Direction.z;

        float tmin = std::max( std::max( std::min(t1, t2), std::min(t3, t4)), std::min(t5, t6));
        float tmax = std::min( std::min( std::max(t1, t2), std::max(t3, t4)), std::max(t5, t6));

        // if tmax < 0, ray is intersecting aabb, but whole aabb is behind us.
        if (tmax < 0)
            return RaycastResult::NoHit();

        // if tmin > tmax, ray doesn't intersect AABB
        if (tmin > tmax)
            return RaycastResult::NoHit();

        float t = 0.f;

        if (tmin < 0.f)
            t = tmax;

        t = tmin;

        Vector3 p0 = this->Origin;
        Vector3 d = this->Direction.Normalize();

        // TODO: Verify this
        Vector3 intersection = p0.Project(d*t);

        // WTF: This algorithm is only valid for spheres!!!
        // TODO: Calculate surfacenormal against rectangle
        Vector3 intersection_from_sphere_origin = (intersection - target.Centroid());

        Vector3 normal = -intersection_from_sphere_origin.Normalize();

        return RaycastResult
        {
            intersection,
            normal,
            true,
            (Shape*)&target
        };
    }

    RaycastResult Ray::Cast(const Plane &target, float maxDistance) {
        float nd = Vector3::Dot(Direction, target.Normal);
        float pn = Vector3::Dot(Origin, target.Normal);

        if (nd >= 0.f)
            return RaycastResult::NoHit();

        float t = (target.distance - pn) / nd;

        Vector3 d = this->Direction.Normalize();

        // TODO: verify this
        Vector3 intersection = this->Origin.Project(d*t);

        // TODO: flip the axis based on direction of incoming ray?
        // Take dot product
        Vector3 normal = target.Normal;

        if (t >= 0.f)
            return RaycastResult
                    {
                        intersection,
                        normal,
                        true,
                        (Shape*) &target
                    };

        return RaycastResult::NoHit();
    }

    Vector3 Ray::ClosestPoint(const Vector3 &targetPoint, float &d) const
    {
        d = std::max(0.f, Vector3::Dot(targetPoint - Origin, Direction));
        return GetPoint(d);
    }

    Vector3 Ray::ClosestPoint(const LineSegment &other, float &d, float &d2) const {
        Line::ClosestPointLineLine(Origin, Direction, other.A, other.B - other.A, d, d2);
        if (d < 0.f)
        {
            d = 0.f;
            if (d2 >= 0.f && d2 <= 1.f)
            {
                other.ClosestPoint(Origin, d2);
                return Origin;
            }

            Vector3 p;
            float t2;

            if (d2 < 0.f)
            {
                p = other.A;
                t2 = 0.f;
            }
            else // u2 > 1.f
            {
                p = other.B;
                t2 = 1.f;
            }

            Vector3 closestPoint = ClosestPoint(p, d);
            Vector3 closestPoint2 = other.ClosestPoint(Origin, d2);
            if (closestPoint.DistanceSquared(p) <= closestPoint2.DistanceSquared(Origin))
            {
                d2 = t2;
                return closestPoint;
            }
            else
            {
                d = 0.f;
                return Origin;
            }
        }
        else if (d2 < 0.f)
        {
            d2 = 0.f;
            return ClosestPoint(other.A, d);
        }
        else if (d2 > 1.f)
        {
            d2 = 1.f;
            return ClosestPoint(other.B, d);
        }
        else
            return GetPoint(d);
    }

    Ray::Ray(const Vector3 &pos, const Vector3 &dir) {
        Origin = pos;
        Direction = dir;
    }

    Vector3 Ray::GetPoint(float distance) const {
        assert(Direction.IsNormalized());
        return Origin + distance * Direction;
    }
}


