#include <J3ML/Geometry/Triangle.h>
#include <J3ML/LinearAlgebra.h>
#include <J3ML/Geometry/AABB.h>
#include <J3ML/Geometry/LineSegment.h>
#include <J3ML/Geometry/Line.h>
#include <J3ML/Geometry/Capsule.h>

namespace J3ML::Geometry
{
    Interval Triangle::ProjectionInterval(const Vector3& axis) const {
        // https://gdbooks.gitbooks.io/3dcollisions/content/Chapter4/generic_sat.html
        float min = axis.Dot(V0);
        float max = min;

        float value = axis.Dot(V1);
        if (value < min)
            min = value;
        if (value > max)
            max = value;

        value = axis.Dot(V2);
        if (value < min)
            min = value;
        if (value > max)
            max = value;
        return Interval{min, max};
    }

    Triangle Triangle::Translated(const Vector3& translation) const {
        return {
            V0 + translation,
            V1 + translation,
            V2 + translation
        };
    }

    Triangle Triangle::Scaled(const Vector3& scale) const {
        return {
            {V0.x * scale.x, V0.y * scale.y, V0.z * scale.y},
            {V1.x * scale.x, V1.y * scale.y, V1.z * scale.y},
            {V2.x * scale.x, V2.y * scale.y, V2.z * scale.y},
        };
    }

    LineSegment Triangle::Edge(int i) const
    {
        assert(0 <= i);
        assert(i <= 2);
        if (i == 0)
            return LineSegment(V0, V1);
        else if (i == 1)
            return LineSegment(V1, V2);
        else if (i == 2)
            return LineSegment(V2, V0);
        else
            return LineSegment(Vector3::NaN, Vector3::NaN);
    }

    Vector3 Triangle::Vertex(int i) const
    {
        assert(0 <= i);
        assert(i <= 2);
        if (i == 0)
            return V0;
        else if (i == 1)
            return V1;
        else if (i == 2)
            return V2;
        else
            return Vector3::NaN;
    }

    Vector3 Triangle::FaceNormal() const {
        Vector3 edge1 = V1 - V0;
        Vector3 edge2 = V2 - V1;

        return edge1.Cross(edge2);
    }

    Plane Triangle::PlaneCCW() const
    {
        return Plane(V0, V1, V2);
    }

    Plane Triangle::PlaneCW() const
    {
        return Plane(V0, V1, V2);
    }

    void Triangle::ProjectToAxis(const Vector3 &axis, float &dMin, float &dMax) const
    {
        dMin = dMax = Vector3::Dot(axis, V0);
        float t = Vector3::Dot(axis, V1);
        dMin = std::min(t, dMin);
        dMax = std::max(t, dMax);
        t = Vector3::Dot(axis, V2);
        dMin = std::min(t, dMin);
        dMax = std::max(t, dMax);
    }

    AABB Triangle::BoundingAABB() const {
        AABB aabb;
        aabb.minPoint = Vector3::Min(V0, V1, V2);
        aabb.maxPoint = Vector3::Max(V0, V1, V2);
        return aabb;
    }

    bool Triangle::Intersects(const AABB &aabb) const {
        return aabb.Intersects(*this);
    }

    /** Calculates the intersection between a line and a triangle. The facing is not accounted for, so
	rays are reported to intersect triangles that are both front and backfacing.
	According to "T. M&ouml;ller, B. Trumbore. Fast, Minimum Storage Ray/Triangle Intersection. 2005."
	http://jgt.akpeters.com/papers/MollerTrumbore97/
	@param linePos The starting point of the line.
	@param lineDir The direction vector of the line. This does not need to be normalized.
	@param v0 Vertex 0 of the triangle.
	@param v1 Vertex 1 of the triangle.
	@param v2 Vertex 2 of the triangle.
	@param u [out] The barycentric u coordinate is returned here if an intersection occurred.
	@param v [out] The barycentric v coordinate is returned here if an intersection occurred.
	@return The distance along the ray to the point of intersection, or +inf if no intersection occurred.
		If no intersection, then u and v and t will contain undefined values. If lineDir was not normalized, then to get the
		real world-space distance, one must scale the returned value with lineDir.Length(). If the returned value is negative,
		then the intersection occurs 'behind' the line starting position, with respect to the direction vector lineDir. */
    float Triangle::IntersectLineTri(const Vector3 &linePos, const Vector3 &lineDir,
                                     const Vector3 &v0, const Vector3 &v1, const Vector3 &v2,
                                     float &u, float &v)
    {
        const float epsilon = 1e-4f;

        // Edge vectors
        Vector3 vE1 = v1 - v0;
        Vector3 vE2 = v2 - v0;

        // begin calculating determinant - also used to calculate U parameter
        Vector3 vP = lineDir.Cross(vE2);

        // If det < 0, intersecting backfacing tri, > 0, intersecting frontfacing tri, 0, parallel to plane.
        const float det = vE1.Dot(vP);

        // If determinant is near zero, ray lies in plane of triangle.
        if (std::abs(det) <= epsilon)
            return INFINITY;
        const float recipDet = 1.f / det;

        // Calculate distance from v0 to ray origin
        Vector3 vT = linePos - v0;

        // Output barycentric u
        u = vT.Dot(vP) * recipDet;
        if (u < -epsilon || u > 1.f + epsilon)
            return INFINITY; // Barycentric U is outside the triangle - early out.

        // Prepare to test V parameter
        Vector3 vQ = vT.Cross(vE1);

        // Output barycentric v
        v = lineDir.Dot(vQ) * recipDet;
        if (v < -epsilon || u + v > 1.f + epsilon) // Barycentric V or the combination of U and V are outside the triangle - no intersection.
            return INFINITY;

        // Barycentric u and v are in limits, the ray intersects the triangle.

        // Output signed distance from ray to triangle.
        return vE2.Dot(vQ) * recipDet;
//	return (det < 0.f) ? IntersectBackface : IntersectFrontface;
    }

    Vector3 Triangle::ClosestPointToTriangleEdge(const Line &other, float *outU, float *outV, float *outD) const
    {
        ///@todo Optimize!
        // The line is parallel to the triangle.
        float unused1, unused2, unused3;
        float d1, d2, d3;
        Vector3 pt1 = Edge(0).ClosestPoint(other, unused1, d1);
        Vector3 pt2 = Edge(1).ClosestPoint(other, unused2, d2);
        Vector3 pt3 = Edge(2).ClosestPoint(other, unused3, d3);
        float dist1 = pt1.DistanceSq(other.GetPoint(d1));
        float dist2 = pt2.DistanceSq(other.GetPoint(d2));
        float dist3 = pt3.DistanceSq(other.GetPoint(d3));
        if (dist1 <= dist2 && dist1 <= dist3)
        {
            if (outU) *outU = BarycentricUV(pt1).x;
            if (outV) *outV = BarycentricUV(pt1).y;
            if (outD) *outD = d1;
            return pt1;
        }
        else if (dist2 <= dist3)
        {
            if (outU) *outU = BarycentricUV(pt2).x;
            if (outV) *outV = BarycentricUV(pt2).y;
            if (outD) *outD = d2;
            return pt2;
        }
        else
        {
            if (outU) *outU = BarycentricUV(pt3).x;
            if (outV) *outV = BarycentricUV(pt3).y;
            if (outD) *outD = d3;
            return pt3;
        }
    }

    Vector3 Triangle::ClosestPointToTriangleEdge(const LineSegment &lineSegment, float *outU, float *outV, float *outD) const
    {
        ///@todo Optimize!
        // The line is parallel to the triangle.
        float unused1, unused2, unused3;
        float d1, d2, d3;
        Vector3 pt1 = Edge(0).ClosestPoint(lineSegment, unused1, d1);
        Vector3 pt2 = Edge(1).ClosestPoint(lineSegment, unused2, d2);
        Vector3 pt3 = Edge(2).ClosestPoint(lineSegment, unused3, d3);
        float dist1 = pt1.DistanceSq(lineSegment.GetPoint(d1));
        float dist2 = pt2.DistanceSq(lineSegment.GetPoint(d2));
        float dist3 = pt3.DistanceSq(lineSegment.GetPoint(d3));
        if (dist1 <= dist2 && dist1 <= dist3)
        {
            if (outU) *outU = BarycentricUV(pt1).x;
            if (outV) *outV = BarycentricUV(pt1).y;
            if (outD) *outD = d1;
            return pt1;
        }
        else if (dist2 <= dist3)
        {
            if (outU) *outU = BarycentricUV(pt2).x;
            if (outV) *outV = BarycentricUV(pt2).y;
            if (outD) *outD = d2;
            return pt2;
        }
        else
        {
            if (outU) *outU = BarycentricUV(pt3).x;
            if (outV) *outV = BarycentricUV(pt3).y;
            if (outD) *outD = d3;
            return pt3;
        }
    }

    Vector3 Triangle::ClosestPoint(const Vector3 &p) const
    {
        /** The code for Triangle-float3 test is from Christer Ericson's Real-Time Collision Detection, pp. 141-142. */

        // Check if P is in vertex region outside A.
        Vector3 ab = V1 - V0;
        Vector3 ac = V2 - V0;
        Vector3 ap = p - V0;
        float d1 = Vector3::Dot(ab, ap);
        float d2 = Vector3::Dot(ac, ap);
        if (d1 <= 0.f && d2 <= 0.f)
            return V0; // Barycentric coordinates are (1,0,0).

        // Check if P is in vertex region outside B.
        Vector3 bp = p - V1;
        float d3 = Vector3::Dot(ab, bp);
        float d4 = Vector3::Dot(ac, bp);
        if (d3 >= 0.f && d4 <= d3)
            return V1; // Barycentric coordinates are (0,1,0).

        // Check if P is in edge region of AB, and if so, return the projection of P onto AB.
        float vc = d1*d4 - d3*d2;
        if (vc <= 0.f && d1 >= 0.f && d3 <= 0.f)
        {
            float v = d1 / (d1 - d3);
            return V0 + v * ab; // The barycentric coordinates are (1-v, v, 0).
        }

        // Check if P is in vertex region outside C.
        Vector3 cp = p - V2;
        float d5 = Vector3::Dot(ab, cp);
        float d6 = Vector3::Dot(ac, cp);
        if (d6 >= 0.f && d5 <= d6)
            return V2; // The barycentric coordinates are (0,0,1).

        // Check if P is in edge region of AC, and if so, return the projection of P onto AC.
        float vb = d5*d2 - d1*d6;
        if (vb <= 0.f && d2 >= 0.f && d6 <= 0.f)
        {
            float w = d2 / (d2 - d6);
            return V0 + w * ac; // The barycentric coordinates are (1-w, 0, w).
        }

        // Check if P is in edge region of BC, and if so, return the projection of P onto BC.
        float va = d3*d6 - d5*d4;
        if (va <= 0.f && d4 - d3 >= 0.f && d5 - d6 >= 0.f)
        {
            float w = (d4 - d3) / (d4 - d3 + d5 - d6);
            return V1 + w * (V2 - V0); // The barycentric coordinates are (0, 1-w, w).
        }

        // P must be inside the face region. Compute the closest point through its barycentric coordinates (u,v,w).
        float denom = 1.f / (va + vb + vc);
        float v = vb * denom;
        float w = vc * denom;
        return V0 + ab * v + ac * w;
    }

    Vector3 Triangle::ClosestPoint(const LineSegment& lineSegment, Vector3 *otherPt) const
    {
        ///\todo Optimize.
        float u, v;
        float t = IntersectLineTri(lineSegment.A, lineSegment.B - lineSegment.A, V0, V1, V2, u, v);
        bool intersects = (t >= 0.0f && t <= 1.0f);
        if (intersects)
        {
//		assume3(lineSegment.GetPoint(t).Equals(this->Point(u, v)), lineSegment.GetPoint(t).SerializeToCodeString(), this->Point(u, v).SerializeToCodeString(), lineSegment.GetPoint(t).Distance(this->Point(u, v)));
            if (otherPt)
                *otherPt = lineSegment.GetPoint(t);
            return this->Point(u, v);
        }

        float u1,v1,d1;
        Vector3 pt1 = ClosestPointToTriangleEdge(lineSegment, &u1, &v1, &d1);

        Vector3 pt2 = ClosestPoint(lineSegment.A);
        Vector3 pt3 = ClosestPoint(lineSegment.B);

        float D1 = pt1.DistanceSq(lineSegment.GetPoint(d1));
        float D2 = pt2.DistanceSq(lineSegment.A);
        float D3 = pt3.DistanceSq(lineSegment.B);

        if (D1 <= D2 && D1 <= D3)
        {
            if (otherPt)
                *otherPt = lineSegment.GetPoint(d1);
            return pt1;
        }
        else if (D2 <= D3)
        {
            if (otherPt)
                *otherPt = lineSegment.A;
            return pt2;
        }
        else
        {
            if (otherPt)
                *otherPt = lineSegment.B;
            return pt3;
        }
    }

    /// Implementation from Christer Ericson's Real-Time Collision Detection, pp. 51-52.
    inline float TriArea2D(float x1, float y1, float x2, float y2, float x3, float y3)
    {
        return (x1-x2)*(y2-y3) - (x2-x3)*(y1-y2);
    }


    Vector3 Triangle::BarycentricUVW(const Vector3 &point) const {
        // Implementation from Christer Ericson's Real-Time Collision Detection, pp. 51-52.

        // Unnormalized triangle normal.
        Vector3 m = Vector3::Cross(V1-V0, V2-V1);

        // Nominators and one-over-denominator for u and v ratios.
        float nu, nv, ood;

        // Absolute components for determining projection plane.
        float x = std::abs(m.x);
        float y = std::abs(m.y);
        float z = std::abs(m.z);

        if (x >= y && x >= z)
        {
            // Project to the yz plane.
            nu = TriArea2D(point.y, point.z, V1.y, V1.z, V2.y, V2.z); // Area of PBC in yz-plane.
            nv = TriArea2D(point.y, point.z, V2.y, V2.z, V0.y, V0.z); // Area OF PCA in yz-plane.
            ood = 1.f / m.x; // 1 / (2*area of ABC in yz plane)
        }
        else if (y >= z) // Note: The book has a redundant 'if (y >= x)' comparison
        {
            // y is largest, project to the xz-plane.
            nu = TriArea2D(point.x, point.z, V1.x, V1.z, V2.x, V2.z);
            nv = TriArea2D(point.x, point.z, V2.x, V2.z, V0.x, V0.z);
            ood = 1.f / -m.y;
        }
        else // z is largest, project to the xy-plane.
        {
            nu = TriArea2D(point.x, point.y, V1.x, V1.y, V2.x, V2.y);
            nv = TriArea2D(point.x, point.y, V2.x, V2.y, V0.x, V0.y);
            ood = 1.f / m.z;
        }
        float u = nu * ood;
        float v = nv * ood;
        float w = 1.f - u - v;
        return Vector3(u,v,w);
    }

    bool Triangle::Intersects(const Capsule &capsule) const {
        return capsule.Intersects(*this);
    }

    namespace {
        bool HaveSeparatingAxis(const Triangle& t1, const Triangle& t2, const Vector3& a, const Vector3& b, const Vector3& c, const Vector3& d) {
            // https://gdbooks.gitbooks.io/3dcollisions/content/Chapter4/robust_sat.html
            Vector3 ab = (a - b);
            Vector3 axis = ab.Cross(c - d);

            if (axis.IsZero()) {
                // Axis is zero, they are parallel, try to find the vector orthogonal to both
                Vector3 n = ab.Cross(c - a);

                if (n.IsZero()) {
                    // AB and AC are parallel, this means they are both on the same axis, just pick one
                    axis = ab;
                } else {
                    // Parallel but not on the same axis, get the vector that is normal to both edges
                    axis = ab.Cross(n);
                }
            }

            return !t1.ProjectionInterval(axis).Intersects(t2.ProjectionInterval(axis));
        }
    }

    bool Intersects(const Triangle& lhs, const Triangle& rhs) {
        // Triangle v Triangle intersection check using SAT
        // https://gdbooks.gitbooks.io/3dcollisions/content/Chapter4/generic_sat.html
        // https://gdbooks.gitbooks.io/3dcollisions/content/Chapter4/triangle-triangle.html
        // Reminder, we only need to find *one* axis to disprove that they collide, the corrolary is that we need to check ALL axes to prove that they collide
        
        Vector3 lEdges[3] = {
            lhs.V1 - lhs.V0,
            lhs.V2 - lhs.V1,
            lhs.V0 - lhs.V2,
        };

        // First, use lhs's face normal as the separating axis
        if (HaveSeparatingAxis(lhs, rhs, lhs.V1, lhs.V0, lhs.V2, lhs.V1)) {
            return false;
        }

        Vector3 rEdges[3] = {
            rhs.V1 - rhs.V0,
            rhs.V2 - rhs.V1,
            rhs.V0 - rhs.V2,
        };

        // Second, use rhs's face normal as the separating axis
        if (HaveSeparatingAxis(lhs, rhs, rhs.V1, rhs.V0, rhs.V2, rhs.V1)) {
            return false;
        }

        // Third, check the normals of each edge against each other
        if (HaveSeparatingAxis(lhs, rhs, lhs.V1, lhs.V0, rhs.V1, rhs.V0) ||
            HaveSeparatingAxis(lhs, rhs, lhs.V1, lhs.V0, rhs.V2, rhs.V1) ||
            HaveSeparatingAxis(lhs, rhs, lhs.V1, lhs.V0, rhs.V0, rhs.V2) ||
            HaveSeparatingAxis(lhs, rhs, lhs.V2, lhs.V1, rhs.V1, rhs.V0) ||
            HaveSeparatingAxis(lhs, rhs, lhs.V2, lhs.V1, rhs.V2, rhs.V1) ||
            HaveSeparatingAxis(lhs, rhs, lhs.V2, lhs.V1, rhs.V0, rhs.V2) ||
            HaveSeparatingAxis(lhs, rhs, lhs.V0, lhs.V2, rhs.V1, rhs.V0) ||
            HaveSeparatingAxis(lhs, rhs, lhs.V0, lhs.V2, rhs.V2, rhs.V1) ||
            HaveSeparatingAxis(lhs, rhs, lhs.V0, lhs.V2, rhs.V0, rhs.V2)) {
            return false;
        }

        // No axis is separating, we can safely conclude they intersect
        return true;
    }

    bool Triangle::Intersects(const Triangle& rhs) const {
        return Geometry::Intersects(*this, rhs);
    }

    Vector3 Triangle::ExtremePoint(const Vector3 &direction) const {
        Vector3 mostExtreme = Vector3::NaN;
        float mostExtremeDist = -FLT_MAX;
        for(int i = 0; i < 3; ++i)
        {
            Vector3 pt = Vertex(i);
            float d = Vector3::Dot(direction, pt);
            if (d > mostExtremeDist)
            {
                mostExtremeDist = d;
                mostExtreme = pt;
            }
        }
        return mostExtreme;
    }

    Vector3 Triangle::ExtremePoint(const Vector3 &direction, float &projectionDistance) const {
        Vector3 extremePoint = ExtremePoint(direction);
        projectionDistance = extremePoint.Dot(direction);
        return extremePoint;
    }

    float Triangle::DistanceSq(const Vector3 &point) const {
        return ClosestPoint(point).DistanceSq(point);
    }

    Vector2 Triangle::BarycentricUV(const Vector3 &point) const {
        Vector3 uvw = BarycentricUVW(point);
        return Vector2{uvw.y, uvw.z};
    }

    Vector3 Triangle::Point(const Vector2 &uv) const {
        return Point(uv.x, uv.y);
    }

    Vector3 Triangle::Point(float u, float v) const {
        // In case the triangle is far away from the origin but is small in size, the elements of 'a' will have large magnitudes,
        // and the elements of (b-a) and (c-a) will be much smaller quantities. Therefore be extra careful with the
        // parentheses and first sum the small floats together before adding it to the large one.
        return V0 + ((V1-V0) * u + (V2-V0) * v);
    }

    Vector3 Triangle::Point(const Vector3 &uvw) const {
        return Point(uvw.x, uvw.y, uvw.z);
    }

    Vector3 Triangle::Point(float u, float v, float w) const {
        return u * V0 + v * V1 + w * V2;
    }

    bool Triangle::Contains(const Vector3 &point, float triangleThicknessSq) const {
        Vector3 normal = (V1-V0).Cross(V2-V0);
        float lenSq = normal.LengthSquared();
        float d = normal.Dot(V1 - point);
        if (d*d > triangleThicknessSq * lenSq)
            return false; ///@todo The plane-point distance test is omitted in Real-Time Collision Detection. p. 25. A bug in the book?

        Vector3 br = BarycentricUVW(point);
        return br.x >= -1e-3f && br.y >= -1e-3f && br.z >= -1e-3f; // Allow for a small epsilon to properly account for points very near the edges of the triangle.
    }
}
