#include <J3ML/LinearAlgebra/Vector3.h>
#include <algorithm>
#include <cassert>
#include <cmath>

namespace J3ML::LinearAlgebra {

    const Vector3 Vector3::Zero     = {0,0,0};
    const Vector3 Vector3::Up       = {0, -1, 0};
    const Vector3 Vector3::Down     = {0, 1, 0};
    const Vector3 Vector3::Left     = {-1, 0, 0};
    const Vector3 Vector3::Right    = {1, 0, 0};
    const Vector3 Vector3::Forward  = {0, 0, -1};
    const Vector3 Vector3::Backward = {0, 0, 1};
    const Vector3 Vector3::NaN      = {NAN, NAN, NAN};
    const Vector3 Vector3::Infinity = {INFINITY, INFINITY, INFINITY};
    const Vector3 Vector3::NegativeInfinity = {-INFINITY, -INFINITY, -INFINITY};

    Vector3 Vector3::operator+(const Vector3& rhs) const
    {
        return {this->x + rhs.x, this->y + rhs.y, this->z + rhs.z};
    }

    Vector3 Vector3::operator-(const Vector3& rhs) const
    {
        return {
                this->x- rhs.x,
                this->y-rhs.y,
                this->z-rhs.z
        };
    }

    Vector3 Vector3::operator*(float rhs) const
    {
        return {
                this->x * rhs,
                this->y * rhs,
                this->z * rhs
        };
    }


    Vector3 Vector3::operator/(float rhs) const
    {
        return {
                this->x / rhs,
                this->y / rhs,
                this->z / rhs
        };
    }

    Vector3 Vector3::operator-() const
    {
        return {-x, -y, -z};
    }




    Vector3::Vector3(): x(0), y(0), z(0) {}

    Vector3::Vector3(float X, float Y, float Z): x(X), y(Y), z(Z) {}

    Vector3::Vector3(const Vector3& rhs) : x(rhs.x), y(rhs.y), z(rhs.z) {}

    Vector3& Vector3::operator=(const Vector3& rhs)
    {
        this->x = rhs.x;
        this->y = rhs.y;
        this->z = rhs.z;
        return *this;
    }

    float Vector3::operator[](std::size_t index) const
    {
        assert(index < 3);
        if (index==0) return x;
        if (index==1) return y;
        if (index==2) return z;
        return 0;
    }

    float &Vector3::operator[](std::size_t index)
    {
        assert(index < 3);
        if (index == 0) return x;
        if (index == 1) return y;
        if (index == 2) return z;
    }

    bool Vector3::IsWithinMarginOfError(const Vector3& rhs, float margin) const
    {
        return this->Distance(rhs) <= margin;
    }

    bool Vector3::operator==(const Vector3& rhs) const
    {
        return this->IsWithinMarginOfError(rhs);
    }

    bool Vector3::operator!=(const Vector3& rhs) const
    {
        return this->IsWithinMarginOfError(rhs) == false;
    }



    Vector3 Vector3::Min(const Vector3& min) const
    {
        return {
                std::min(this->x, min.x),
                std::min(this->y, min.y),
                std::min(this->z, min.z)
        };
    }

    Vector3 Vector3::Max(const Vector3& max) const
    {
        return {
                std::max(this->x, max.x),
                std::max(this->y, max.y),
                std::max(this->z, max.z)
        };
    }

    Vector3 Vector3::Clamp(const Vector3& min, const Vector3& max) const
    {
        return {
                std::clamp(this->x, min.x, max.x),
                std::clamp(this->y, min.y, max.y),
                std::clamp(this->z, min.z, max.z)
        };
    }

    float Vector3::Distance(const Vector3& to) const
    {
        return ((*this)-to).Magnitude();
    }

    float Vector3::Length() const
    {
        return std::sqrt(LengthSquared());
    }

    float Vector3::LengthSquared() const
    {
        return (x*x + y*y + z*z);
    }

    float Vector3::Magnitude() const
    {
        return std::sqrt(x*x + y*y + z*z);
    }

    float Vector3::Dot(const Vector3& rhs) const
    {
        return x * rhs.x +
               y * rhs.y +
               z * rhs.z;
    }

    Vector3 Vector3::Project(const Vector3& rhs) const
    {
        float scalar = this->Dot(rhs) / (rhs.Magnitude()*rhs.Magnitude());
        return rhs * scalar;
    }

    Vector3 Vector3::Cross(const Vector3& rhs) const
    {
        return {
                this->y * rhs.z - this->z * rhs.y,
                this->z * rhs.x - this->x * rhs.z,
                this->x * rhs.y - this->y * rhs.x
        };
    }

    Vector3 Vector3::Normalize() const
    {
        if (Length() > 0)
            return {
                    x / Length(),
                    y / Length(),
                    z / Length()
            };
        else
            return {0,0,0};
    }

    Vector3 Vector3::Lerp(const Vector3& goal, float alpha) const
    {
        return this->operator*(1.0f - alpha) + (goal * alpha);
    }

    float Vector3::GetX() const { return x;}

    float Vector3::GetY() const { return y;}

    float Vector3::GetZ() const { return z;}

    void Vector3::SetX(float newX) { x = newX;}

    void Vector3::SetY(float newY) { y = newY;}

    void Vector3::SetZ(float newZ) { z = newZ;}

    Vector3 Vector3::Min(const Vector3 &lhs, const Vector3 &rhs) {
        return lhs.Min(rhs);
    }

    Vector3 Vector3::Max(const Vector3 &lhs, const Vector3 &rhs) {
        return lhs.Max(rhs);
    }

    Vector3 Vector3::Clamp(const Vector3 &min, const Vector3 &input, const Vector3 &max) {
        return input.Clamp(min, max);
    }

    float Vector3::Distance(const Vector3 &from, const Vector3 &to) {
        return from.Distance(to);
    }

    float Vector3::Length(const Vector3 &of) {
        return of.Length();
    }

    float Vector3::LengthSquared(const Vector3 &of) {
        return of.LengthSquared();
    }

    bool Vector3::IsPerpendicular(const Vector3 &other, float epsilonSq) const {
        float dot = Dot(other);
        return dot*dot <= epsilonSq * LengthSquared() * other.LengthSquared();
    }

    bool Vector3::IsZero(float epsilonSq) const {
        return LengthSquared() <= epsilonSq;
    }

    bool Vector3::IsNormalized(float epsilonSq) const {
        return std::abs(LengthSquared()-1.f) <= epsilonSq;
    }

    Vector3 Vector3::Cross(const Vector3 &lhs, const Vector3 &rhs) {
        return lhs.Cross(rhs);
    }

    Vector3 Vector3::Normalize(const Vector3 &targ) {
        return targ.Normalize();
    }

    Vector3 Vector3::Project(const Vector3 &lhs, const Vector3 &rhs) {
        return lhs.Project(rhs);
    }

    float Vector3::Dot(const Vector3 &lhs, const Vector3 &rhs) {
        return lhs.Dot(rhs);
    }

    float Vector3::Magnitude(const Vector3 &of) {
        return of.Magnitude();
    }

    Vector3 Vector3::Lerp(const Vector3 &lhs, const Vector3 &rhs, float alpha) {
        return lhs.Lerp(rhs, alpha);
    }

    Vector3 Vector3::Add(const Vector3 &lhs, const Vector3 &rhs) {
        return lhs.Add(rhs);
    }

    Vector3 Vector3::Add(const Vector3 &rhs) const {
        return *this + rhs;
    }

    Vector3 Vector3::Sub(const Vector3 &rhs) const {
        return *this - rhs;
    }

    Vector3 Vector3::Sub(const Vector3 &lhs, const Vector3 &rhs) {
        return lhs.Sub(rhs);
    }

    Vector3 Vector3::Mul(float scalar) const {
        return *this * scalar;
    }

    Vector3 Vector3::Mul(const Vector3 &lhs, float rhs) {
        return lhs.Mul(rhs);
    }

    Vector3 Vector3::Div(float scalar) const {
        return *this / scalar;
    }

    Vector3 Vector3::Div(const Vector3 &lhs, float rhs) {
        return lhs.Div(rhs);
    }

    Angle2D Vector3::AngleBetween(const Vector3 &rhs) const {
        const auto Pi_x_180 = 180.f / M_PI;
        auto dist = this->Distance(rhs);
        float x = -(asinf((rhs.y - this->y) / dist));
        float y = (atan2f(rhs.x - this->x,rhs.z - this->z));
        return {x, y};
    }

    Angle2D Vector3::AngleBetween(const Vector3 &lhs, const Vector3 &rhs) // TODO: 3D Angle representation?
    {
        return lhs.AngleBetween(rhs);
    }

    Vector3 Vector3::Direction(const Vector3 &rhs) {
        float x = (cos(Math::Radians(rhs.y)) * cos(Math::Radians(rhs.x)));
        float y = -sin(Math::Radians(rhs.x));
        float z = (sin(Math::Radians(rhs.y)) * cos(Math::Radians(rhs.x)));
        return {x, y, z};
    }

    Vector3 Vector3::Abs() const {
        return {std::abs(x), std::abs(y), std::abs(z)};
    }

    float *Vector3::ptr() {
        return &x;
    }

    void Vector3::Orthonormalize(Vector3 &a, Vector3 &b) {
        a = a.Normalize();
        b = b - b.ProjectToNorm(a);
        b = b.Normalize();
    }

    void Vector3::Orthonormalize(Vector3 &a, Vector3 &b, Vector3 &c) {
        a = a.Normalize();
        b = b - b.ProjectToNorm(a);
        b = b.Normalize();
        c = c - c.ProjectToNorm(a);
        c = c - c.ProjectToNorm(b);
        c = c.Normalize();
    }

    Vector3 Vector3::ProjectToNorm(const Vector3 &direction) const {
        return direction * this->Dot(direction);
    }

    bool Vector3::IsFinite() const {
        return std::isfinite(x) && std::isfinite(y) && std::isfinite(z);
    }

    Vector3::Vector3(const float *data) {
        x = data[0];
        y = data[1];
        z = data[2];
    }

    Vector3 &Vector3::operator+=(const Vector3 &rhs) {
        x += rhs.x;
        y += rhs.y;
        z += rhs.z;
        return *this;
    }

    Vector3 &Vector3::operator-=(const Vector3 &rhs) {
        x -= rhs.x;
        y -= rhs.y;
        z -= rhs.z;
        return *this;
    }

    Vector3 &Vector3::operator*=(float scalar) {
        x *= scalar;
        y *= scalar;
        z *= scalar;
        return *this;
    }

    Vector3 &Vector3::operator/=(float scalar) {
        x /= scalar;
        y /= scalar;
        z /= scalar;
        return *this;
    }

    float Vector3::MinElement() const {
        return std::min(x, std::min(y, z));
    }

    bool Vector3::AreOrthonormal(const Vector3 &a, const Vector3 &b, float epsilon) {
        return a.IsPerpendicular(b, epsilon) && a.IsNormalized(epsilon*epsilon) && b.IsNormalized(epsilon*epsilon);
    }

    Vector3 Vector3::Mul(const Vector3 &rhs) const {
        return {
                this->x * rhs.x,
                this->y * rhs.y,
                this->z * rhs.z
        };
    }

    Vector3 Vector3::Div(const Vector3 &v) const {
        return {
                this->x/v.x,
                this->y/v.y,
                this->z/v.z
        };
    }

    Vector3 Vector3::Abs(const Vector3 &rhs) {
        return rhs.Abs();
    }

    bool Vector3::Equals(const Vector3 &rhs, float epsilon) const {
        return std::abs(x - rhs.x) < epsilon &&
               std::abs(y - rhs.y) < epsilon &&
               std::abs(z - rhs.z) < epsilon;
    }

    bool Vector3::Equals(float _x, float _y, float _z, float epsilon) const {
        return std::abs(x - _x) < epsilon &&
               std::abs(y - _y) < epsilon &&
               std::abs(z - _z) < epsilon;
    }

    Vector3 Vector3::Min(const Vector3 &a, const Vector3 &b, const Vector3 &c) {
        return {
                std::min(a.x, std::min(b.x, c.x)),
                std::min(a.y, std::min(b.y, c.y)),
                std::min(a.z, std::min(b.z, c.z))
        };
    }

    Vector3 Vector3::Max(const Vector3 &a, const Vector3 &b, const Vector3 &c) {
        return {
                std::max(a.x, std::max(b.x, c.x)),
                std::max(a.y, std::max(b.y, c.y)),
                std::max(a.z, std::max(b.z, c.z))
        };
    }

    Vector3 Vector3::Perpendicular(const Vector3 &hint, const Vector3 &hint2) const {
        assert(!this->IsZero());
        assert(hint.IsNormalized());
        assert(hint2.IsNormalized());
        Vector3 v = this->Cross(hint);
        float len = v.TryNormalize();

    }

    float Vector3::TryNormalize() {
        assert(IsFinite());
        float length = Length();
        if (length > 1e-6f)
        {
            *this *= 1.f / length;
            return length;
        }
        else
        {
            Set(1.f, 0.f, 0.f); // We will always produce a normalized vector.
            return 0; // But signal failure, so user knows we have generated an arbitrary normalization.
        }
    }

    float Vector3::At(int index) const {
        assert(index >= 0);
        assert(index < Dimensions);
        return ptr()[index];
    }

    float &Vector3::At(int index) {
        assert(index >= 0);
        assert(index < Dimensions);
        return ptr()[index];
    }

    void Vector3::Set(float d, float d1, float d2) {
        x = d;
        y = d1;
        z = d2;
    }

    Vector3::Vector3(const Vector2& XY, float Z) {
        x = XY.x;
        y = XY.y;
        z = Z;
    }

    Vector3::Vector3(float scalar) {
        x = scalar;
        y = scalar;
        z = scalar;
    }

    float Vector3::DistanceSquared(const Vector3 &to) const {
        return (*this-to).LengthSquared();
    }

    float Vector3::DistanceSquared(const Vector3 &from, const Vector3 &to) {
        return from.DistanceSquared(to);
    }


}