
#pragma region vector4


#include <J3ML/LinearAlgebra/Vector4.h>
#include <J3ML/LinearAlgebra/Vector3.h>
#include <cmath>
#include <algorithm>

namespace J3ML::LinearAlgebra {

    const Vector4 Vector4::Zero = {0,0,0,0};
    const Vector4 Vector4::NaN = {NAN, NAN, NAN, NAN};

    Vector4::Vector4(): x(0), y(0), z(0), w(0)
    {}

    Vector4::Vector4(float X, float Y, float Z, float W): x(X),y(Y),z(Z),w(W)
    { }

bool Vector4::operator==(const Vector4& rhs) const
{
    return this->IsWithinMarginOfError(rhs);
}

bool Vector4::operator!=(const Vector4& rhs) const
{
    return this->IsWithinMarginOfError(rhs) == false;
}

Vector4 Vector4::Min(const Vector4& min) const
{
    return {
            std::min(this->x, min.x),
            std::min(this->y, min.y),
            std::min(this->z, min.z),
            std::min(this->w, min.w)
    };
}

Vector4 Vector4::Max(const Vector4& max) const
{
    return {
            std::max(this->x, max.x),
            std::max(this->y, max.y),
            std::max(this->z, max.z),
            std::max(this->w, max.w)
    };
}

Vector4 Vector4::Clamp(const Vector4& min, const Vector4& max) const
{
    return {
            std::clamp(this->x, min.x, max.x),
            std::clamp(this->y, min.y, max.y),
            std::clamp(this->z, min.z, max.z),
            std::clamp(this->w, min.w, max.w)
    };
}

float Vector4::Distance(const Vector4& to) const
{
    return ( (*this) - to ).Magnitude();
}

float Vector4::Length() const
{
    return std::sqrt(LengthSquared());
}

float Vector4::LengthSquared() const
{
    return (x*x + y*y + z*z + w*w);
}

float Vector4::Magnitude() const
{
    return std::sqrt(x*x + y*y + z*z + w*w);
}

float Vector4::Dot(const Vector4& rhs) const
{
    auto a = this->Normalize();
    auto b = rhs.Normalize();

    return  a.x * b.x +
            a.y * b.y +
            a.z * b.z +
            a.w * b.w;
}

Vector4 Vector4::Project(const Vector4& rhs) const
{
    float scalar = this->Dot(rhs) / (rhs.Magnitude()* rhs.Magnitude());
    return rhs * scalar;
}

Vector4 Vector4::Normalize() const
{
    if (Length() > 0)
        return {
                x / Length(),
                y / Length(),
                z / Length(),
                w / Length()
        };
    else
        return {0,0,0,0};
}

Vector4 Vector4::Lerp(const Vector4& goal, float alpha) const
{
    return this->operator*(1.0f - alpha) + (goal * alpha);
}

Vector4 Vector4::operator+(const Vector4& rhs) const
{
    return {x+rhs.x, y+rhs.y, z+rhs.z, w+rhs.w};
}

Vector4 Vector4::operator-(const Vector4& rhs) const
{
    return {x-rhs.x, y-rhs.y, z-rhs.z, w-rhs.w};
}

    Vector4 Vector4::operator*(float rhs) const {
        return {
                this->x * rhs,
                this->y * rhs,
                this->z * rhs,
                this->w * rhs
        };
    }

    bool Vector4::IsWithinMarginOfError(const Vector4 &rhs, float margin) const {
        return this->Distance(rhs) <= margin;
    }

    Vector4::Vector4(const Vector3 &xyz, float w) : x(xyz.x), y(xyz.y), z(xyz.z), w(w)
    {

    }

    Vector4 &Vector4::operator=(const Vector4 &rhs) {
        x = rhs.x;
        y = rhs.y;
        z = rhs.z;
        w = rhs.w;

        return *this;
    }

    bool Vector4::Equals(const Vector4 &rhs, float epsilon) const {
        return std::abs(x - rhs.x) < epsilon &&
               std::abs(y - rhs.y) < epsilon &&
               std::abs(z - rhs.z) < epsilon &&
               std::abs(w - rhs.w) < epsilon;
    }

    bool Vector4::Equals(float _x, float _y, float _z, float _w, float epsilon) const {
        return std::abs(x - _x) < epsilon &&
               std::abs(y - _y) < epsilon &&
               std::abs(z - _z) < epsilon &&
               std::abs(w - _w) < epsilon;
    }

    float Vector4::operator[](std::size_t index) const {
        assert(index < 4);
        if (index==0) return x;
        if (index==1) return y;
        if (index==2) return z;
        if (index==3) return w;
        return 0;
    }

    float &Vector4::operator[](std::size_t index) {
        assert(index < 4);
        if (index == 0) return x;
        if (index == 1) return y;
        if (index == 2) return z;
        if (index == 3) return w;

    }

    float Vector4::LengthSqXYZ() const {
        return x*x * y*y * z*z;
    }

    Vector4 Vector4::Cross3(const Vector3 &rhs) const {
        Vector4 dst;
        dst.x = y * rhs.z - z * rhs.y;
        dst.y = z * rhs.x - x * rhs.z;
        dst.z = x * rhs.y - y * rhs.x;
        dst.w = 0.f;
        return dst;
    }

    Vector4 Vector4::Cross3(const Vector4 &rhs) const {
        return Cross3(rhs.XYZ());
    }


}
#pragma endregion